/*
 * QVM86 virtualisation module for qemu
 *
 * Copyright (C) 2005 Filip Navara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Windows kernel interfacing code.
 */

#include "qvm86.h"

void *
host_alloc_zeroed(unsigned long size)
{
    void * ptr;

    ptr = ExAllocatePoolWithTag(NonPagedPool, size, TAG('Q','V','M','Z'));
    if (!ptr)
        return NULL;
    RtlZeroMemory(ptr, size);
    return ptr;
}

void *
host_alloc_pages(int count)
{
    char * ptr;

    ptr = MmAllocateNonCachedMemory(PAGE_SIZE * count);
    if (!ptr)
        return NULL;
    RtlZeroMemory(ptr, PAGE_SIZE * count);
    return ptr;
}

void
host_free(void *ptr)
{
    if (!ptr)
        return;
    ExFreePool(ptr);
}

void
host_free_pages(void *ptr, int count)
{
    if (!ptr)
        return;
    MmFreeNonCachedMemory(ptr, PAGE_SIZE * count);
}

unsigned long
host_phys_addr(void * addr)
{
    LARGE_INTEGER pa;
    pa = MmGetPhysicalAddress(addr);
    return pa.LowPart;
}

void
host_map_user(unsigned long useraddr, qvm86_mapped_page *map)
{
    PMDL mdl;
    PPFN_NUMBER mdl_pages;

    /* fail if the user address isn't aligned */
    if (useraddr & 0xfff)
    {
        DbgPrint("qvm86: Tried to map unaligned user memory (%p)\n", useraddr);
        map->host = NULL;
        return;
    }

    mdl = ExAllocatePool(NonPagedPool, sizeof(MDL) + sizeof(PFN_NUMBER));
    if (mdl == NULL)
    {
        DbgPrint("qvm86: Not enough memory for MDL structure\n");
        map->host = NULL;
        return;
    }
    mdl_pages = (PPFN_NUMBER)(mdl + 1);

    MmInitializeMdl(mdl, useraddr, PAGE_SIZE);
    MmProbeAndLockPages(mdl, KernelMode, IoModifyAccess);

    map->host = MmGetSystemAddressForMdlSafe(mdl, NormalPagePriority);
    if (map->host == NULL)
    {
        DbgPrint("qvm86: Can't map MDL to kernel address space\n");
        ExFreePool(mdl);
        return;
    }
    map->hostdata = mdl;
    map->phys = mdl_pages[0] << PAGE_SHIFT;
    map->refcount = 1;
}

void
host_unmap_user(qvm86_mapped_page *map, int dirty)
{
    PMDL mdl = map->hostdata;

    /* FIXME: Should we do something about the dirty flag? */

    MmUnmapLockedPages(map->host, mdl);
    MmUnlockPages(mdl);
    ExFreePool(mdl);

    /* TODO: This should probably be done elsewhere.  */
    map->host = NULL;
    map->phys = 0;
}

void
host_mark_dirty(qvm86_mapped_page *map)
{
    /* FIXME */
}

NTSTATUS STDCALL
KQemuCreate(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);
    qvm86_state *State;

    State = host_alloc_zeroed(sizeof(qvm86_state));
    if (State == NULL)
    {
        Irp->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
        Irp->IoStatus.Information = 0;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    IrpStack->FileObject->FsContext = State;
    qvm86_init_device(State);
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;
}

NTSTATUS STDCALL
KQemuClose(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);
    qvm86_state *State = IrpStack->FileObject->FsContext;

    qvm86_cleanup_device(State);
    host_free(State);

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;
}

NTSTATUS STDCALL
KQemuDeviceControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);
    qvm86_state *State = IrpStack->FileObject->FsContext;
    NTSTATUS Status;
    int ret;

    Irp->IoStatus.Information = 0;

    switch (IrpStack->Parameters.DeviceIoControl.IoControlCode)
    {
        case KQEMU_INIT:
            if (IrpStack->Parameters.DeviceIoControl.InputBufferLength <
                sizeof(struct kqemu_init)) 
            {
                Status = STATUS_INVALID_PARAMETER;
                break;
            }

            qvm86_init(State, Irp->AssociatedIrp.SystemBuffer);
            Status = STATUS_SUCCESS;
            break;

        case KQEMU_EXEC:
            {
                struct kqemu_cpu_state s;

                if (IrpStack->Parameters.DeviceIoControl.InputBufferLength <
                    sizeof(s) ||
                    IrpStack->Parameters.DeviceIoControl.OutputBufferLength <
                    sizeof(s))
                {
                    Status = STATUS_INVALID_PARAMETER;
                    break;
                }
            
                RtlCopyMemory(&s, Irp->AssociatedIrp.SystemBuffer, sizeof(s));
                ret = qvm86_exec(State, &s);
                s.retval = ret;
                RtlCopyMemory(Irp->AssociatedIrp.SystemBuffer, &s, sizeof(s));
                Irp->IoStatus.Information = sizeof(s);
                Status = STATUS_SUCCESS;
            }
            break;
    
        case KQEMU_GET_VERSION:
            if (IrpStack->Parameters.DeviceIoControl.OutputBufferLength <
                sizeof(int)) 
            {
                Status = STATUS_INVALID_PARAMETER;
                break;
            }

            *((int *)Irp->AssociatedIrp.SystemBuffer) = KQEMU_VERSION;
            Irp->IoStatus.Information = sizeof(int);
            Status = STATUS_SUCCESS;
            break;

        default:
            Status = STATUS_INVALID_PARAMETER;
            break;
    }

    Irp->IoStatus.Status = Status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return Status;
}

VOID STDCALL
KQemuUnload(PDRIVER_OBJECT DriverObject)
{
    IoDeleteDevice(DriverObject->DeviceObject);
}

NTSTATUS STDCALL 
DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath)
{
    PDEVICE_OBJECT DeviceObject;
    UNICODE_STRING DeviceName;
    UNICODE_STRING SymlinkName;
    NTSTATUS Status;

    DbgPrint("QVM86 Accelerator Module version %d.%d.%d\n",
             (KQEMU_VERSION >> 16),
             (KQEMU_VERSION >> 8) & 0xff,
             (KQEMU_VERSION) & 0xff);

    MmLockPagableCodeSection(DriverEntry);

    DriverObject->MajorFunction[IRP_MJ_CREATE] = KQemuCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = KQemuClose;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = KQemuDeviceControl;
    DriverObject->DriverUnload = KQemuUnload;

    RtlInitUnicodeString(&DeviceName, L"\\Device\\qvm86");
    RtlInitUnicodeString(&SymlinkName, L"\\??\\qvm86");

    Status = IoCreateDevice(DriverObject, 0,
                            &DeviceName, FILE_DEVICE_UNKNOWN, 0, FALSE,
                            &DeviceObject);
    if (!NT_SUCCESS(Status))
        return Status;

    /* Create the dos device link */
    IoCreateSymbolicLink(&SymlinkName, &DeviceName);

    return STATUS_SUCCESS;
}
