/* QPlex x86 virtualisation module for qemu

   Copyright (C) 2005 Paul Brook

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Header file with declarations for everything.
 */
/* TODO: Split this into qemu-host.h and qemu-monitor.h.  */

#ifdef _WIN32
#include "qvm86-win32.h"
#else
#include <linux/linkage.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#endif

#include "kqemu.h"

#define FIELD_S(s, o, n) uint16_t s##_##n;
#define FIELD_L(s, o, n) uint32_t s##_##n;
#define FIELD_END(o)
typedef struct {
#include "monitor-env.h"
} qvm86_monitor_env;

/* These bits are private.  */
/* TODO: The following comment is out of date and inaccurate.  */
/* We have four and-a-bit VMs:
    user_: Host usermode.
    host_: Host kernel.
    
    monitor_: Privileged monitor stub.
    guest_: The virtualised guest.
    shadow_: The monitor-side user pagetables used to fake the guest VM.

   Virtualisation works as follows:
   Host kernel allocates a monitor gdt, and copies the in-use guest entries
   into it.  Monitor segments are added in free entries.  We currently use
   the same pagetables in both host kernel and monitor space. This probably
   needs to change so that we can cope with guest/host PAE and amd64.

   Guest task descriptor is copied and a monitor task created in spare
   monitor gdt slot.

   We can't use the guest pagetables directly, so setup a shadow pagetable.
   Guest traps are passed back to the host kernel via the monitor.
   Pagefaults occur for two reasons:
    1) A missing entry in the shadow pagetable.  We page-in the required
       page, fill in the shadow pagetable and resume guest execution.
    2) A real guest Pagefault.  Execution stops and we pass the exception
       back to userspace.
   Other guest traps are also passed back to userspace.
   When guest execution  finishes we flush the shadow pagetable accessed
   and dirty bits back to the real guest pagetables.
   We currently reset the shadow pagetable every time we are called from
   userspace.
 */

typedef struct
{
  uint32_t low;
  uint32_t high;
} qvm86_gdt_entry;

/* The contents of a TSS segment.  */
typedef struct
{
  uint16_t prev;
  uint16_t pad0;
  uint32_t esp0;
  uint16_t ss0;
  uint16_t pad1;
  uint32_t esp1;
  uint16_t ss1;
  uint16_t pad2;
  uint32_t esp2;
  uint16_t ss2;
  uint16_t pad3;
  uint32_t cr3;
  uint32_t eip;

  uint32_t eflags;
  uint32_t eax;
  uint32_t ecx;
  uint32_t edx;
  uint32_t ebx;
  uint32_t esp;
  uint32_t ebp;
  uint32_t esi;
  uint32_t edi;

  uint16_t es;
  uint16_t pad4;
  uint16_t cs;
  uint16_t pad5;
  uint16_t ss;
  uint16_t pad6;
  uint16_t ds;
  uint16_t pad7;
  uint16_t fs;
  uint16_t pad8;
  uint16_t gs;
  uint16_t pad9;
  uint16_t ldt;
  uint16_t pad10;
  
  uint16_t T:1;
  uint16_t pad11:15;
  uint16_t iomap;
} qvm86_tss;

/* A page of guest memory.  */
typedef struct
{
  uint32_t *host;
  
  uint32_t phys;  /* TODO: This field can be removed.  */
  /* We used to keep track of pte entries for pages so the we could swap them
     back out.  Need to figure out how to do this when a page is referenced
     by multiple pagetable entries.  This happens eg. with the "zero" page
     under linux.  */
  /* TODO: This will be totally bogus once we implement host side
     page fault handling.  */
  int refcount;
  //uint32_t *pte;
  void * hostdata;
} qvm86_mapped_page;

/* A shadow PDE.  */
typedef struct
{
  /* The shadow pagetable.  */
  uint32_t *virt;
  /* The physical address of the shadow pagetable.  */
  uint32_t phys;
  /* The guest pagetable we are shadowing.  */
  qvm86_mapped_page *guest_pte;
} qvm86_shadow_pde;

/* Stub used for trap handlers where the CPU already pushed an error code.  */
typedef struct
{
  uint8_t  pushl;    /* Always 0x68 == pushl            */
  uint32_t vector;   /* Interrupt vector number         */
  uint8_t  jmp;      /* Always 0xe9 == jmp              */
  uint32_t reloc;    /* Relative offset of destination  */
} __attribute__ ((packed)) idt_method1_t;

/* Stub used for trap handlers where the CPU did not push an error code.  */
typedef struct
{
  uint8_t  pushla;   /* Always 0x68 == pushl            */
  uint32_t dummy;    /* Dummy error code                */
  uint8_t  pushlb;   /* Always 0x68 == pushl            */
  uint32_t vector;   /* Interrupt vector number         */
  uint8_t  jmp;      /* Always 0xe9 == jmp              */
  uint32_t reloc;    /* Relative offset of destination  */
} __attribute__ ((packed)) idt_method2_t;

typedef union
{
  idt_method1_t m1;
  idt_method2_t m2;
} qvm86_stub;

#define IDT_INT		0 /* Interrupt.  */
#define IDT_EFAULT	1 /* Fault with error code.  */
#define IDT_FAULT	2 /* Fault without error code.  */

typedef void asmlinkage (*qvm86_monitorproc)(qvm86_monitor_env *);

typedef struct
{
  qvm86_monitor_env * monitor_env;
  qvm86_gdt_entry * monitor_gdt;
  qvm86_gdt_entry * monitor_idt;
  qvm86_stub *monitor_stubs;
  /* Points to the top of the monitor stack.  */
  uint32_t *monitor_stack;
  /* The task state structure for the monitor task.  */
  qvm86_tss * monitor_tss;
  /* The shadow page directory.  */ 
  uint32_t *shadow_pgd;
  /* A copy of the guest page direcory.  */
  uint32_t *pgd_copy;
  /* The host addresses of the shadow pagetables.  */
  qvm86_shadow_pde *shadow_pde;
  uint32_t guest_cr3;
  uint32_t monitor_cr3;
  /* The pagetable entries for monitor space.  */
  uint32_t *monitor_pagetable;
  uint32_t monitor_pagetable_physaddr;
  uint32_t *switch_pagetable;
  qvm86_mapped_page *guest_pgd;

  /* Pointer to guest->host physical page mappings.  */
  uint32_t **monitor_physmem_index;
  /* Monitor pagetable entries for guest->host physical page mappings.  */
  uint32_t *monitor_physmem_pte;
  uint32_t monitor_physmem_pte_physaddr;
  /* Monitor pagetable entries to map the shadow pagetables into monitor
     space.  */
  uint32_t *monitor_shadow_pte;
  uint32_t monitor_shadow_pte_physaddr;

  /* The host userspace address of guest physical memory.  */
  char * user_physmem;
  unsigned long user_physmem_size;
  /* Emulator IO memory mapping.  */
  uint32_t *l1_phys_map;
  /* Emulator dirty memory flags.  */
  uint8_t *phys_ram_dirty;
  unsigned long *pages_to_flush;
  /* Satus of mapped-in host pages.  */
  qvm86_mapped_page *host_physmem_map;

  /* This points to the shared copy, and ensures the code doesn't disappear
     when we switch to monitor space.  */
  qvm86_monitorproc monitor_run;

  /* This is the location of the monitor space.  It may be moved around.  */
  uint32_t monitor_base;

  /* maddr_* are offsets from monitor_base.  */
  uint32_t maddr_handleint;
  uint32_t maddr_handlefault;
  uint32_t maddr_start_guest;

  uint32_t maddr_stubs; /* monitor_stubs */
  uint32_t maddr_gdt; /* monitor_gdt */
  uint32_t maddr_idt; /* monitor_idt */
  uint32_t maddr_stack; /* monitor_stack */
  uint32_t maddr_tss; /* monitor_tss */
  /* The monitor pde for mapping host space while switching.  */
  uint32_t maddr_switch_ptr;

  /* This will be nonzero if a fatal error occured.  Usually due to
     something not yet implemented.  Ideally this would never happen.  */
  int die;
} qvm86_state;

void qvm86_init (qvm86_state * qs, struct kqemu_init * init);
int qvm86_exec (qvm86_state * qs, struct kqemu_cpu_state * hostenv);
void qvm86_init_device (qvm86_state *);
void qvm86_cleanup_device (qvm86_state *);
void * host_alloc_pages (int);
void * host_alloc_zeroed (unsigned long);
void host_free (void *);
void host_free_pages (void *, int);
void host_map_user (unsigned long, qvm86_mapped_page *);
void host_unmap_user (qvm86_mapped_page *, int);
void host_mark_dirty (qvm86_mapped_page *);
unsigned long host_phys_addr (void *);

/* in switch.S */
void asmlinkage qvm86_monitor_run(qvm86_monitor_env *);

/* Monitor-side routines.  */
void qvm86_monitor_return (void);
void qvm86_monitor_entry (void);
void qvm86_start_guest (void);
void qvm86_monitor_switch (void);
void qvm86_monitor_inthandler (void);
void qvm86_monitor_faulthandler (void);
